# stdlib
from math import nan

schema_dict = {
    "project": {
        0: "example-project",
        1: "example-project",
        2: "example-project",
        3: "example-project",
        4: "example-project",
        5: "example-project",
        6: "example-project",
        7: "example-project",
        8: "example-project",
        9: "example-project",
        10: "example-project",
        11: "example-project",
        12: "example-project",
        13: "example-project",
        14: "example-project",
        15: "example-project",
        16: "example-project",
        17: "example-project",
        18: "example-project",
        19: "example-project",
        20: "example-project",
        21: "example-project",
        22: "example-project",
    },
    "dataset_id": {
        0: "test_1gb",
        1: "test_1gb",
        2: "test_1gb",
        3: "test_1gb",
        4: "test_1gb",
        5: "test_1gb",
        6: "test_1gb",
        7: "test_1gb",
        8: "test_1gb",
        9: "test_1gb",
        10: "test_1gb",
        11: "test_1gb",
        12: "test_1gb",
        13: "test_1gb",
        14: "test_1gb",
        15: "test_1gb",
        16: "test_1gb",
        17: "test_1gb",
        18: "test_1gb",
        19: "test_1gb",
        20: "test_1gb",
        21: "test_1gb",
        22: "test_1gb",
    },
    "table_id": {
        0: "posts",
        1: "posts",
        2: "posts",
        3: "posts",
        4: "posts",
        5: "posts",
        6: "posts",
        7: "comments",
        8: "comments",
        9: "comments",
        10: "comments",
        11: "comments",
        12: "comments",
        13: "comments",
        14: "comments",
        15: "comments",
        16: "comments",
        17: "comments",
        18: "comments",
        19: "comments",
        20: "comments",
        21: "comments",
        22: "comments",
    },
    "schema_name": {
        0: "int64_field_0",
        1: "id",
        2: "name",
        3: "subscribers_count",
        4: "permalink",
        5: "nsfw",
        6: "spam",
        7: "int64_field_0",
        8: "id",
        9: "body",
        10: "parent_id",
        11: "created_at",
        12: "last_modified_at",
        13: "gilded",
        14: "permalink",
        15: "score",
        16: "comment_id",
        17: "post_id",
        18: "author_id",
        19: "spam",
        20: "deleted",
        21: "upvote_raio",
        22: "collapsed_in_crowd_control",
    },
    "schema_field": {
        0: "INTEGER",
        1: "STRING",
        2: "STRING",
        3: "INTEGER",
        4: "STRING",
        5: "FLOAT",
        6: "BOOLEAN",
        7: "INTEGER",
        8: "STRING",
        9: "STRING",
        10: "STRING",
        11: "INTEGER",
        12: "INTEGER",
        13: "BOOLEAN",
        14: "STRING",
        15: "INTEGER",
        16: "STRING",
        17: "STRING",
        18: "STRING",
        19: "BOOLEAN",
        20: "BOOLEAN",
        21: "FLOAT",
        22: "BOOLEAN",
    },
    "description": {
        0: "None",
        1: "None",
        2: "None",
        3: "None",
        4: "None",
        5: "None",
        6: "None",
        7: "None",
        8: "None",
        9: "None",
        10: "None",
        11: "None",
        12: "None",
        13: "None",
        14: "None",
        15: "None",
        16: "None",
        17: "None",
        18: "None",
        19: "None",
        20: "None",
        21: "None",
        22: "None",
    },
    "num_rows": {
        0: "2000000",
        1: "2000000",
        2: "2000000",
        3: "2000000",
        4: "2000000",
        5: "2000000",
        6: "2000000",
        7: "2000000",
        8: "2000000",
        9: "2000000",
        10: "2000000",
        11: "2000000",
        12: "2000000",
        13: "2000000",
        14: "2000000",
        15: "2000000",
        16: "2000000",
        17: "2000000",
        18: "2000000",
        19: "2000000",
        20: "2000000",
        21: "2000000",
        22: "2000000",
    },
}


query_dict = {
    "int64_field_0": {
        0: 4,
        1: 5,
        2: 10,
        3: 16,
        4: 17,
        5: 23,
        6: 24,
        7: 25,
        8: 27,
        9: 40,
    },
    "id": {
        0: "t5_via1x",
        1: "t5_cv9gn",
        2: "t5_8p2tq",
        3: "t5_8fcro",
        4: "t5_td5of",
        5: "t5_z01fv",
        6: "t5_hmqjk",
        7: "t5_1flyj",
        8: "t5_5rwej",
        9: "t5_uurcv",
    },
    "name": {
        0: "/channel/mylittlepony",
        1: "/channel/polyamory",
        2: "/channel/Catholicism",
        3: "/channel/cordcutters",
        4: "/channel/stevenuniverse",
        5: "/channel/entitledbitch",
        6: "/channel/engineering",
        7: "/channel/nottheonion",
        8: "/channel/FoodPorn",
        9: "/channel/puppysmiles",
    },
    "subscribers_count": {
        0: 4323081,
        1: 2425929,
        2: 4062607,
        3: 7543226,
        4: 2692168,
        5: 2709080,
        6: 8766144,
        7: 2580984,
        8: 7784809,
        9: 3715991,
    },
    "permalink": {
        0: "/channel//channel/mylittlepony",
        1: "/channel//channel/polyamory",
        2: "/channel//channel/Catholicism",
        3: "/channel//channel/cordcutters",
        4: "/channel//channel/stevenuniverse",
        5: "/channel//channel/entitledbitch",
        6: "/channel//channel/engineering",
        7: "/channel//channel/nottheonion",
        8: "/channel//channel/FoodPorn",
        9: "/channel//channel/puppysmiles",
    },
    "nsfw": {
        0: nan,
        1: nan,
        2: nan,
        3: nan,
        4: nan,
        5: nan,
        6: nan,
        7: nan,
        8: nan,
        9: nan,
    },
    "spam": {
        0: False,
        1: False,
        2: False,
        3: False,
        4: False,
        5: False,
        6: False,
        7: False,
        8: False,
        9: False,
    },
}
