# relative
class KeyValueBackingStore:
    pass


class KeyValueStorePartition:
    """Key-Value StorePartition
    Parameters:
        `settings`: PartitionSettings
            PySyft specific settings
        `store_config`: StoreConfig
            Backend specific configuration
    """

    pass
